
#include "kxen.h"

#define KXEN_DEFINE_SYMBOLS_PROTO
#include <kxen_link.h>
KXEN_PROTOTYPES(extern)

unsigned long kxen_cpu_vm = 0;

#if defined(__x86_64__)
#define to_affinity(x) ((ULONGLONG)1 << (x))
#else
#define to_affinity(x) (1 << (x))
#endif

unsigned long
kxen_first_cpu(void)
{
    KAFFINITY affinity;
    unsigned long host_cpu;

    affinity = KeQueryActiveProcessors();
    for (host_cpu = 0; host_cpu < MAXIMUM_PROCESSORS; host_cpu++) {
	if (affinity & to_affinity(host_cpu))
	    break;
    }
    ASSERT(host_cpu != MAXIMUM_PROCESSORS);
    return host_cpu;
}

void
kxen_cpu_pin(unsigned long host_cpu)
{
    KAFFINITY affinity;

    affinity = to_affinity(host_cpu);
    KeSetSystemAffinityThread(affinity);
}

void
kxen_cpu_pin_current(void)
{
    unsigned long cpu;

    cpu = KeGetCurrentProcessorNumber();
    kxen_cpu_pin(cpu);
}

void
kxen_cpu_pin_first(void)
{
    unsigned long cpu;

    cpu = kxen_first_cpu();
    kxen_cpu_pin(cpu);
}

void
kxen_cpu_unpin(void)
{

    KeRevertToUserAffinityThread();
}

void
kxen_set_cpu_active_mask(void *mask, int mask_size)
{
    KAFFINITY affinity;
    unsigned long host_cpu;

    affinity = KeQueryActiveProcessors();
    ASSERT(sizeof(affinity) <= mask_size);
    memcpy(mask, &affinity, sizeof(affinity));

    for (host_cpu = kxen_first_cpu() + 1; host_cpu < MAXIMUM_PROCESSORS;
	 host_cpu++) {
	if (affinity & to_affinity(host_cpu))
	    break;
    }
    ASSERT(host_cpu != MAXIMUM_PROCESSORS);
    kxen_cpu_vm = host_cpu;
}

void __cdecl
kxen_on_each_cpu(void (*fn)(void *), void *arg)
{
    KAFFINITY affinity;
    CCHAR host_cpu;

    affinity = KeQueryActiveProcessors();
    for (host_cpu = MAXIMUM_PROCESSORS - 1; host_cpu >= 0; host_cpu--) {
	if ((affinity & to_affinity(host_cpu)) == 0)
	    continue;
	kxen_cpu_pin(host_cpu);
	dprintk("kxen_on_each_cpu on cpu %d: %p(%p)\n", host_cpu, fn, arg);
	try {
	    fn(arg);
	} except (kxen_except_handler(GetExceptionCode(),
				      GetExceptionInformation())) {
	}
    }
    kxen_cpu_unpin();
}
